'use strict'

const jsonld = require('jsonld')
const jsig = require('jsonld-signatures')
const bitcore = require('bitcore-lib')
const Web3 = require('web3')
const ethUtil = require('ethereumjs-util');

jsig.use('jsonld', jsonld)

const SSIF_SIGNATURE_ALGORITHM = 'EcdsaKoblitzSignature2016'
const SSIF_BLOCKCHAIN_ALGORITHM_ERC2018 = 'SsifErc780Hash2018'

const SSIF_COMMITMENT_PREDICATE = 'commitment'

var TRUSTED_KEYS = {}
var TRUSTED_CONTRACTS = []
var TRUSTED_ISSUERS = []

var ETHEREUM_CONTRACT_ADDRESS = '0x72d3bc4da6e025540edbfee8b376918f95ec59e4';
var ETHEREUM_PROVIDER = 'https://rinkeby.infura.io';

module.exports = {
    addTrustedContract: addTrustedContract,
    addTrustedPublicKey: addTrustedPublicKey,
    addTrustedIssuer: addTrustedIssuer,
    setEthereumProvider: setEthereumProvider,
    getContractAt: getContractAt, // Debug
    createAttestation: createAttestation,
    formatPublicKey: formatPublicKey,
    signECDSAAttestation: signECDSAAttestation,
    verifyECDSAAttestation: verifyECDSAAttestation, // Debug
    verifyAttestation: verifyAttestation,
    verifyAttestations: verifyAttestations,
    TRUSTED_KEYS: TRUSTED_KEYS // Debug
}

// const SSIF_CONTEXT_URL = 'http://tinyurl.com/ssifschema-1'

function addTrustedPublicKey(publicKey) {
    if (!publicKey) {
        return false
    }

    TRUSTED_KEYS[publicKey] = formatPublicKey(publicKey)

    return TRUSTED_KEYS[publicKey]
}

function addTrustedContract(address) {
    if (!address) {
        return false
    }

    TRUSTED_CONTRACTS.push(address)

    return address
}

function addTrustedIssuer(address) {
    if (!address) {
        return false
    }

    TRUSTED_ISSUERS.push(address.toLowerCase())

    return address
}

function setEthereumProvider(provider) {
    if (!provider) {
        return false
    }

    ETHEREUM_PROVIDER = provider

    return provider
}

function formatPublicKey(publicKey) {
    return {
        '@context': jsig.SECURITY_CONTEXT_URL,
        id: 'ecdsa-koblitz-pubkey:' + publicKey,
        type: 'CryptographicKey',
        publicKeyWif: publicKey
    }
}

function getTrustedPublicKey(publicKey, options = {}) {
    const key = publicKey.replace(/^ecdsa-koblitz-pubkey:/, '')
    const trustedKeys = options.trustedPublicKeys || TRUSTED_KEYS
    return trustedKeys[key]
}

function getTrustedContract(contract, options = {}) {
    const address = contract.replace(/^ethereum-contract:/, '')
    const trustedContracts = options.trustedContracts || TRUSTED_CONTRACTS

    if (trustedContracts.indexOf(address) !== -1) {
        return address
    }

    return null
}

function getTrustedIssuer(issuer, options = {}) {
    const address = issuer.replace(/^ethereum-address:/, '').toLowerCase()
    const trustedIssuers = options.trustedIssuers || TRUSTED_ISSUERS

    if (trustedIssuers.indexOf(address) !== -1) {
        return address
    }

    return null
}


async function verifyAttestations(attestations, options = {}) {
    return Promise.all(attestations.map(function (attestation) {
        return verifyAttestation(attestation, options)
    }))
}

async function verifyAttestation(attestation, options = {}) {
    if (attestation.commitment) {
        if (attestation.commitment.type === SSIF_SIGNATURE_ALGORITHM) {
            return verifyECDSAAttestation(attestation, options)
        } else if (attestation.commitment.type === SSIF_BLOCKCHAIN_ALGORITHM_ERC2018) {
            return verifyBlockchainAttestationErc780(attestation, options)
        }
    }

    throw new Error('Attestation cannot be validated')
}

async function verifyECDSAAttestation(attestation, options = {}) {
    const copy = Object.assign({}, attestation)

    const commitment = copy.commitment
    delete copy.commitment

    copy['signature'] = commitment

    const publicKey = getTrustedPublicKey(commitment.creator, {
        trustedPublicKeys: options.trustedPublicKeys
    })

    if (!publicKey) {
        throw new Error('Not created by trusted public key')
    }

    const publicKeyOwner = {
        '@context': jsig.SECURITY_CONTEXT_URL,
        publicKey: [publicKey]
    }

    if (await jsig.promises.verify(copy, Object.assign({
        publicKey: publicKey,
        publicKeyOwner: publicKeyOwner
    }, options))) {
        return attestation
    }

    throw new Error('Not valid')
}

/**
 * used for SSIFEthereumHash2017
 *
 * @param {*} issuer
 * @param {*} subject
 * @param {*} predicate
 * @param {*} object
 */
async function verifyBlockchainStatementErc780(issuer, subject, predicate, object) {
    let erc780Contract = await getContractAt(ETHEREUM_CONTRACT_ADDRESS)
    let value = await erc780Contract.methods.getClaim(
        issuer,
        hash(subject),
        hash(subject + predicate)).call()
    return (value.slice(0, 42) === hash(subject + object));
}

/**
 * Used for SSIFEthereumErc780Hash2018
 *
 * @param {*} attestation
 * @param {*} options
 */
async function verifyBlockchainAttestationErc780(attestation, options = {}) {
    const issuer = getTrustedIssuer(attestation.commitment.issuer, options)

    if (!issuer) {
        throw new Error('Not a trusted issuer')
    }

    const statements = Object.assign({}, attestation.statements)

    const address = statements['@id'].replace(/^ecdsa-koblitz-pubkey:/, '')
    const publicKeyString = attestation.commitment['publicKey'].replace(/^(ecdsa-koblitz-pubkey:)?(0x)?/, '')

    delete statements['@id']

    statements['@context'] = attestation['@context']

    console.log(statements)

    const publicKey = bitcore.PublicKey(publicKeyString).toAddress('livenet')

    if (publicKey.toString() !== address) {
        throw new Error(`Subject is not the address of ${publicKey}`)
    }

    const compacted = await jsonld.promises.compact(statements, {})

    for (let predicate in compacted) {
        var object = compacted[predicate]
        var objectValue = object

        if (object instanceof Object && '@value' in object) {
            objectValue = object['@value']
        }

        // Dirty fix, just make it work for now :-(
        const originalPredicate = predicate.replace(/#/, '/');

        const result = await verifyBlockchainStatementErc780(
            issuer,
            '0x' + publicKeyString,
            originalPredicate,
            objectValue
        )
        console.log("RESULT originalPredicate:", result)
        if (!result) throw new Error(`Attestation for ${publicKeyString} ${originalPredicate} ${compacted[predicate]} invalid`);
    }

    return attestation
}


function createAttestation(subject, predicate, object) {
    let statements = {}

    if (typeof subject === 'undefined') {
        throw new Error('Need to pass a subject or statements')
    }

    if (typeof predicate === 'undefined' && typeof object === 'undefined') {
        statements = subject
    } else {
        if (subject !== null) {
            statements['@id'] = subject
        }
        statements[predicate] = object
    }

    return {
        '@context': [
            'http://schema.org/',
            'https://w3id.org/security/v1',
            'https://w3id.org/identity/v1',
            'http://tinyurl.com/ssifschema-1'
        ],
        '@type': 'Attestation',
        statements: statements
    }
}

async function signECDSAAttestation(attestation, privateKey, options = {}) {
    // -> Use privateKey + jsig to sign statements.
    // -> change 'signature' to 'commitment'
    const addressFromPrivateKey = new bitcore.PrivateKey(privateKey).toPublicKey().toAddress('livenet')

    const signed = await jsig.promises.sign(attestation, Object.assign({
        privateKeyWif: privateKey,
        algorithm: SSIF_SIGNATURE_ALGORITHM,
        creator: 'ecdsa-koblitz-pubkey:' + addressFromPrivateKey
    }, options))

    const commitment = signed.signature
    delete signed.signature

    signed[SSIF_COMMITMENT_PREDICATE] = commitment

    return signed
}

const CONTRACT_ABI = [{
    'constant': false,
    'inputs': [{
        'name': 'key',
        'type': 'bytes32'
    },
        {
            'name': 'value',
            'type': 'bytes32'
        }
    ],
    'name': 'setSelfClaim',
    'outputs': [],
    'payable': false,
    'stateMutability': 'nonpayable',
    'type': 'function'
},
    {
        'constant': false,
        'inputs': [{
            'name': 'subject',
            'type': 'address'
        },
            {
                'name': 'key',
                'type': 'bytes32'
            },
            {
                'name': 'value',
                'type': 'bytes32'
            }
        ],
        'name': 'setClaim',
        'outputs': [],
        'payable': false,
        'stateMutability': 'nonpayable',
        'type': 'function'
    },
    {
        'constant': false,
        'inputs': [{
            'name': 'issuer',
            'type': 'address'
        },
            {
                'name': 'subject',
                'type': 'address'
            },
            {
                'name': 'key',
                'type': 'bytes32'
            }
        ],
        'name': 'removeClaim',
        'outputs': [],
        'payable': false,
        'stateMutability': 'nonpayable',
        'type': 'function'
    },
    {
        'constant': true,
        'inputs': [{
            'name': 'issuer',
            'type': 'address'
        },
            {
                'name': 'subject',
                'type': 'address'
            },
            {
                'name': 'key',
                'type': 'bytes32'
            }
        ],
        'name': 'getClaim',
        'outputs': [{
            'name': '',
            'type': 'bytes32'
        }],
        'payable': false,
        'stateMutability': 'view',
        'type': 'function'
    },
    {
        'constant': true,
        'inputs': [{
            'name': '',
            'type': 'address'
        },
            {
                'name': '',
                'type': 'address'
            },
            {
                'name': '',
                'type': 'bytes32'
            }
        ],
        'name': 'registry',
        'outputs': [{
            'name': '',
            'type': 'bytes32'
        }],
        'payable': false,
        'stateMutability': 'view',
        'type': 'function'
    },
    {
        'anonymous': false,
        'inputs': [{
            'indexed': true,
            'name': 'issuer',
            'type': 'address'
        },
            {
                'indexed': true,
                'name': 'subject',
                'type': 'address'
            },
            {
                'indexed': true,
                'name': 'key',
                'type': 'bytes32'
            },
            {
                'indexed': false,
                'name': 'value',
                'type': 'bytes32'
            },
            {
                'indexed': false,
                'name': 'updatedAt',
                'type': 'uint256'
            }
        ],
        'name': 'ClaimSet',
        'type': 'event'
    },
    {
        'anonymous': false,
        'inputs': [{
            'indexed': true,
            'name': 'issuer',
            'type': 'address'
        },
            {
                'indexed': true,
                'name': 'subject',
                'type': 'address'
            },
            {
                'indexed': true,
                'name': 'key',
                'type': 'bytes32'
            },
            {
                'indexed': false,
                'name': 'removedAt',
                'type': 'uint256'
            }
        ],
        'name': 'ClaimRemoved',
        'type': 'event'
    }
]

var CONTRACT, web3

async function getWeb3() {
    // Try to catch a lost connection -- not sure if this works as intended

    if (!web3) {
        web3 = new Web3(Web3.givenProvider || ETHEREUM_PROVIDER);
    }

    await web3.eth.net.isListening().catch(() => {
        web3 = new Web3(Web3.givenProvider || ETHEREUM_PROVIDER);
    });

    return web3
}


async function getContractAt(address) {
    address = address.replace(/^ethereum-contract:/, '')

    if (!CONTRACT) {
        let web3 = await getWeb3();
        CONTRACT = new web3.eth.Contract(CONTRACT_ABI, address);
    }

    return CONTRACT;
}

function hash(data) {
    return '0x' + ethUtil.sha3(data).slice(-20).toString('hex');
}

/*

  Speed up JSON LD

*/

var _defaults = {
    headers: {
        accept: 'application/ld+json, application/json'
    }
};

jsonld.documentLoaders.nodeCached = function (options) {
    options = options || {};
    var headers = buildHeaders(options.headers);
    var strictSSL = ('strictSSL' in options) ? options.strictSSL : true;
    var maxRedirects = ('maxRedirects' in options) ? options.maxRedirects : -1;
    var request = ('request' in options) ? options.request : require('request');
    var http = require('http');
    // TODO: disable cache until HTTP caching implemented
    var cache = new jsonld.DocumentCache();

    var queue = new jsonld.RequestQueue();
    if (options.usePromise) {
        return queue.wrapLoader(function (url) {
            return jsonld.promisify(loadDocument, url, []);
        });
    }

    return queue.wrapLoader(function (url, callback) {
        loadDocument(url, [], callback);
    });

    function loadDocument(url, redirects, callback) {
        if (url.indexOf('http:') !== 0 && url.indexOf('https:') !== 0) {
            return callback(new JsonLdError(
                'URL could not be dereferenced; only "http" and "https" URLs are ' +
                'supported.',
                'jsonld.InvalidUrl', {
                    code: 'loading document failed',
                    url: url
                }), {
                contextUrl: null,
                documentUrl: url,
                document: null
            });
        }
        if (options.secure && url.indexOf('https') !== 0) {
            return callback(new JsonLdError(
                'URL could not be dereferenced; secure mode is enabled and ' +
                'the URL\'s scheme is not "https".',
                'jsonld.InvalidUrl', {
                    code: 'loading document failed',
                    url: url
                }), {
                contextUrl: null,
                documentUrl: url,
                document: null
            });
        }
        // TODO: disable cache until HTTP caching implemented
        var doc = cache.get(url);
        if (doc !== null) {
            return callback(null, doc);
        }

        request({
            url: url,
            headers: headers,
            strictSSL: strictSSL,
            followRedirect: false
        }, handleResponse);

        function handleResponse(err, res, body) {
            doc = {
                contextUrl: null,
                documentUrl: url,
                document: body || null
            };

            // handle error
            if (err) {
                return callback(new JsonLdError(
                    'URL could not be dereferenced, an error occurred.',
                    'jsonld.LoadDocumentError', {
                        code: 'loading document failed',
                        url: url,
                        cause: err
                    }), doc);
            }
            var statusText = http.STATUS_CODES[res.statusCode];
            if (res.statusCode >= 400) {
                return callback(new JsonLdError(
                    'URL could not be dereferenced: ' + statusText,
                    'jsonld.InvalidUrl', {
                        code: 'loading document failed',
                        url: url,
                        httpStatusCode: res.statusCode
                    }), doc);
            }

            // handle Link Header
            if (res.headers.link &&
                res.headers['content-type'] !== 'application/ld+json') {
                // only 1 related link header permitted
                var linkHeader = jsonld.parseLinkHeader(
                    res.headers.link)[LINK_HEADER_REL];
                if (_isArray(linkHeader)) {
                    return callback(new JsonLdError(
                        'URL could not be dereferenced, it has more than one associated ' +
                        'HTTP Link Header.',
                        'jsonld.InvalidUrl', {
                            code: 'multiple context link headers',
                            url: url
                        }), doc);
                }
                if (linkHeader) {
                    doc.contextUrl = linkHeader.target;
                }
            }

            // handle redirect
            if (res.statusCode >= 300 && res.statusCode < 400 &&
                res.headers.location) {
                if (redirects.length === maxRedirects) {
                    return callback(new JsonLdError(
                        'URL could not be dereferenced; there were too many redirects.',
                        'jsonld.TooManyRedirects', {
                            code: 'loading document failed',
                            url: url,
                            httpStatusCode: res.statusCode,
                            redirects: redirects
                        }), doc);
                }
                if (redirects.indexOf(url) !== -1) {
                    return callback(new JsonLdError(
                        'URL could not be dereferenced; infinite redirection was detected.',
                        'jsonld.InfiniteRedirectDetected', {
                            code: 'recursive context inclusion',
                            url: url,
                            httpStatusCode: res.statusCode,
                            redirects: redirects
                        }), doc);
                }
                redirects.push(url);
                return loadDocument(res.headers.location, redirects, callback);
            }
            // cache for each redirected URL
            redirects.push(url);
            // TODO: disable cache until HTTP caching implemented
            for (var i = 0; i < redirects.length; ++i) {
                cache.set(
                    redirects[i], {
                        contextUrl: null,
                        documentUrl: redirects[i],
                        document: body
                    });
            }
            callback(err, doc);
        }
    }
};

function buildHeaders(optionsHeaders) {
    optionsHeaders = optionsHeaders || {};

    var hasAccept = Object.keys(optionsHeaders).some(function (h) {
        return h.toLowerCase() === 'accept';
    });

    if (hasAccept) {
        throw new RangeError(
            'Accept header may not be specified as an option; only "' +
            _defaults.headers.accept + '" is supported.');
    }

    var headers = {
        'Accept': _defaults.headers.accept
    };
    for (var k in optionsHeaders) {
        headers[k] = optionsHeaders[k];
    }

    return headers;
}

jsonld.useDocumentLoader('nodeCached')